# std
import logging
import os
import sys
import traceback
from typing import Optional as Opt, Union

# constants
FMT = '%(asctime)s [%(levelname)s] %(message)s'
ENVIRONMENT_LOG_LEVEL = os.environ.get('LOG_LEVEL', '').upper()

# loggers
_loggers = {}


def get_logger(
        name: str = 'laz',
        level: Opt[Union[int, str]] = None,
        verbosity: Opt[int] = None,
        default_level: Opt[Union[int, str]] = None,
) -> logging.Logger:
    if name in _loggers:
        logger = _loggers[name]
    else:
        logger = _get_logger(name, default_level)
        _loggers[name] = logger

    if level is not None:
        if isinstance(level, str):
            logger.setLevel(level.upper())
        else:
            logger.setLevel(level)

    if verbosity is not None and verbosity > 0:
        logger.setLevel(_convert_verbosity_to_level(verbosity))

    return logger


def _get_logger(
        name: str,
        default_level: Opt[Union[int, str]] = None,
) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.propagate = False

    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(logging.Formatter(fmt=FMT))
    logger.addHandler(handler)

    if default_level is not None and isinstance(default_level, str):
        default_level = default_level.upper()

    try:
        logger.setLevel(ENVIRONMENT_LOG_LEVEL or default_level or 'WARNING')
    except ValueError as e:
        logger.warning(f'Unknown log level: {ENVIRONMENT_LOG_LEVEL}\n{e}\n{traceback.format_exc()}')
    return logger


def _convert_verbosity_to_level(verbosity: int) -> int:
    if verbosity <= 0:
        return logging.WARNING
    elif verbosity == 1:
        return logging.INFO
    elif verbosity >= 2:
        return logging.DEBUG
    elif verbosity >= 3:
        return 5
    elif verbosity >= 4:
        return 1
    else:
        return logging.WARNING




# std
import argparse
import traceback
from typing import List

# internal
from laz.utils.errors import LazRuntimeError
from laz.model.tree import Node
from laz.model.path import Path
from laz.model.resolver import Resolver
from laz.model.target import Target
from laz.model.act import Act


class Runner:

    def __init__(self, root_node: Node, cli_args: argparse.Namespace, args: List[str]):
        self.root_node = root_node
        self.cli_args = cli_args
        self.path = Path(args[0])
        self.args = args[1:]

    def resolve(self) -> List[Target]:
        resolver = Resolver(self.root_node, self.path)
        targets = resolver.resolve()
        if self.cli_args.reverse:
            targets.reverse()
        return targets

    def run(self):
        failures = []
        targets = self.resolve()
        for target in targets:
            try:
                act = Act(target, self.args)
                act.act()
            except Exception as e:
                failures.append((e, traceback.format_tb()))
        for failure in failures:
            print(failure)
        if failures:
            raise LazRuntimeError('Some actions failed')



# std
from __future__ import annotations
from typing import List, Optional as Opt

# internal
from laz.model.configuration import Configuration
from laz.utils.funcs import flatten


class Node:
    configuration: Configuration
    parent: Opt[Node]
    children: List[Node]

    def __init__(
            self,
            configuration: Configuration,
            parent: Opt[Node] = None,
            children: Opt[List[Node]] = None,
    ):
        self.configuration = configuration
        self.parent = parent
        self.children = children or []

    @property
    def is_root(self):
        return self.parent is None

    @property
    def is_leaf(self):
        return len(self.children) == 0

    def add_child(self, node: Node):
        node.parent = self
        self.children.append(node)

    def add_children(self, nodes: List[Node]):
        for node in nodes:
            self.add_child(node)

    def nodes(self) -> List[Node]:
        if self.is_leaf:
            return [self]
        else:
            return [self] + flatten([n.nodes() for n in self.children])

    def root_path(self) -> List[Node]:
        if self.is_root:
            return [self]
        else:
            return self.parent.root_path() + [self]



# std
from typing import List

# internal
from laz.utils.contexts import in_dir
from laz.model.action import Action
from laz.model.target import Target
from laz.utils.templates import evaluate


class Act:

    def __init__(self, target: Target, args: List[str]):
        target.push({'args': args})
        target.push(evaluate(target.data))
        self.target = target
        self.args = evaluate(args, target.data)
        self.action = Action.new(self.target, ' '.join(self.args))

    def act(self):
        with in_dir(self.target.data['dirpath']):
            self.action.run()



# std
from __future__ import annotations


class Serializable:

    def serialize(self) -> str:
        raise NotImplementedError

    @classmethod
    def deserialize(cls, name: str, serialized: str) -> Serializable:
        raise NotImplementedError



# std
from __future__ import annotations
from typing import (
    Generic,
    Iterable,
    Iterator as _Iterator,
    Optional as Opt,
    Sized,
    TypeVar,
    Union,
)

# types
T = TypeVar('T')


class Iterator(Generic[T]):

    def __init__(self, iterable: Union[Iterable[T], Sized]):
        self._iterator: _Iterator[T] = iter(iterable)
        self._index: int = -1
        self._length: int = len(iterable)
        self._value: Opt[T] = None

    def __iter__(self):
        return self

    def __next__(self) -> Iterator:
        self._index += 1
        self._value = next(self._iterator)
        return self

    @property
    def is_first(self) -> bool:
        return self._index == 0

    @property
    def is_last(self) -> bool:
        return self._index == self._length - 1

    @property
    def value(self) -> T:
        return self._value



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['laz',
 'laz.cli',
 'laz.cli.subcommands',
 'laz.internal',
 'laz.model',
 'laz.plugins',
 'laz.utils']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3,<4', 'PyYAML>=6,<7', 'poetry>=1.1.13,<2.0.0']

entry_points = \
{'console_scripts': ['laz = laz.main:main']}

setup_kwargs = {
    'name': 'laz',
    'version': '0.0.1a5',
    'description': 'CLI tool to configure and run parameterized actions against targets.',
    'long_description': '# Laz\n\nA CLI tool to configure and run parameterized actions against targets.\n\n## Installation\n\nLaz currently requires Python 3.8+.\n\n```shell\npip3 install --user laz\n```\n\nYou can check your installation by running:\n\n```shell\nlaz version\n```\n',
    'author': 'Josh Wycuff',
    'author_email': 'Josh.Wycuff@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



# std
from __future__ import annotations
from typing import List

# internal
from laz.utils.types import DictData
from laz.utils.merge import merge


class Stackable:

    _stack: List[DictData] = None
    data: DictData = None

    def push(self, data: DictData):
        if self._stack is None:
            self._stack = []
        self._stack.append(self.data)
        self.data = merge(self.data, data)

    def pop(self):
        self.data = self._stack.pop()



# internal
from laz.utils.logging import get_logger
from laz.cli.parser import parser
from laz.utils.load import load
from laz.model.tree import Node
from laz.utils.iterator import Iterator
from laz.utils.node import get_node_configuration


subparsers = parser.add_subparsers(required=True)

parser_tree = subparsers.add_parser('tree')
parser_tree.add_argument('--show-targets', '-t', action='store_true', default=False, help='Show targets.')


def tree():
    args = parser.parse_args()
    get_logger(verbosity=args.verbose)
    root = load()
    print(f'{root.configuration.name} {_targets(root, args.show_targets)}')
    for child in Iterator(root.children):
        _tree(child.value, '', child.is_last, args.show_targets)


def _tree(node: Node, prefix: str = '', last: bool = False, show_targets: bool = False):
    symbol = '└─' if last else '├─'
    print(f'{prefix}{symbol} {node.configuration.name} {_targets(node, show_targets)}')

    for child in Iterator(node.children):
        _tree(child.value, prefix + '│  ', child.is_last, show_targets)


def _targets(node: Node, show_targets: bool = False) -> str:
    if show_targets:
        configuration = get_node_configuration(node)
        if len(configuration.target_names) > 0:
            names = ', '.join(configuration.target_names)
            return f'({names})'
    return ''
